----------------------------------------------------------
--		Image Script				--
--							--
--  Author:	Nathaniel Meyer				--
--  Email:	nate@nutty.ca				--
--  Copyright:	Nutty Productions			--
--  Website:	http://www.nutty.ca			--
--							--
--  Desc:	Contains a set of image routines	--
--		Dst = background image			--
--		Src = layer/blending image		--
--							--
----------------------------------------------------------


--[[
	Colour
		@return: A colour structure with (r,g,b,a) values
]]
function Colour (r, g, b, a)
	return { r = r,
		 g = g,
		 b = b,
		 a = a };
end

function ColourCopy ( colour )
	return { r = colour.r,
		 g = colour.g,
		 b = colour.b,
		 a = colour.a };
end

function ColourFloat ( colour )
	colour.r = colour.r / 255;
	colour.g = colour.g / 255;
	colour.b = colour.b / 255;

	return colour;
end

function ColourInvert ( colour )
	colour.r = 1 - (colour.r / 255);
	colour.g = 1 - (colour.g / 255);
	colour.b = 1 - (colour.b / 255);

	return colour;
end


--[[
	imgColourCheck
		@desc: Map colour values from [0, 255]
]]
function imgColourCheck (colour)
	-- Check lower boundary
	colour.r = math.max(colour.r, 0);
	colour.g = math.max(colour.g, 0);
	colour.b = math.max(colour.b, 0);

	-- Check upper boundary
	colour.r = math.min(colour.r, 255);
	colour.g = math.min(colour.g, 255);
	colour.b = math.min(colour.b, 255);

	return colour;
end


--[=[
	Mix Blends
]=]
--[[
	imgMix
		@return: Blended Value
		@desc: (1-a)src + (a)dst
]]
function imgMix (dst, src, blend)
	blend.r = (src.r * (1.0 - dst.a)) + (dst.r * dst.a);
	blend.g = (src.g * (1.0 - dst.a)) + (dst.g * dst.a);
	blend.b = (src.b * (1.0 - dst.a)) + (dst.b * dst.a);

	return imgColourCheck(blend);
end

--[=[
	Darken Blends
]=]
--[[
	imgDarken
		@return: Blended Value
		@desc: min(dst, src)
]]
function imgDarken (dst, src, blend)
	blend.r = math.min(dst.r, src.r);
	blend.g = math.min(dst.g, src.g);
	blend.b = math.min(dst.b, src.b);

	return imgColourCheck(blend);
end

--[[
	imgMultiply
		@return: Blended Value
		@desc: dst * src
]]
function imgMultiply (dst, src, blend)
	blend.r = (dst.r * src.r) / 255;
	blend.g = (dst.g * src.g) / 255;
	blend.b = (dst.b * src.b) / 255;

	return imgColourCheck(blend);
end

--[[
	imgColourBurn
		@return: Blended Value
		@desc: 1 - (1 - dst) / src
]]
function imgColourBurn (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	blend.r = (1 - (1 - dst.r) / src.r) * 255;
	blend.g = (1 - (1 - dst.g) / src.g) * 255;
	blend.b = (1 - (1 - dst.b) / src.b) * 255;

	return imgColourCheck(blend);
end

--[[
	imgLinearBurn
		@return: Blended Value
		@desc: dst + src - 1
]]
function imgLinearBurn (dst, src, blend)
	blend.r = dst.r + src.r - 255;
	blend.g = dst.g + src.g - 255;
	blend.b = dst.b + src.b - 255;

	return imgColourCheck(blend);
end


--[=[
	Lighten Blends
]=]
--[[
	imgLighten
		@return: Blended Value
		@desc: max(dst, src)
]]
function imgLighten (dst, src, blend)
	blend.r = math.max(dst.r, src.r);
	blend.g = math.max(dst.g, src.g);
	blend.b = math.max(dst.b, src.b);

	return imgColourCheck(blend);
end

--[[
	imgScreen
		@return: Blended Value
		@desc: 1 - (1 - dst) * (1 - src)
]]
function imgScreen (dst, src, blend)
	-- Invert the colours
	ColourInvert(dst);
	ColourInvert(src);

	-- Calculate the blend
	blend.r = (1 - (dst.r * src.r)) * 255;
	blend.g = (1 - (dst.g * src.g)) * 255;
	blend.b = (1 - (dst.b * src.b)) * 255;

	return imgColourCheck(blend);
end

--[[
	imgColourDodge
		@return: Blended Value
		@desc: dst / (1 - src)
]]
function imgColourDodge (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	blend.r = (dst.r / (1 - src.r)) * 255;
	blend.g = (dst.g / (1 - src.g)) * 255;
	blend.b = (dst.b / (1 - src.b)) * 255;

	return imgColourCheck(blend);
end

--[[
	imgLinearDodge
		@return: Blended Value
		@desc: dst + src
]]
function imgLinearDodge (dst, src, blend)
	blend.r = dst.r + src.r;
	blend.g = dst.g + src.g;
	blend.b = dst.b + src.b;

	return imgColourCheck(blend);
end


--[=[
	Mix Blends
]=]
--[[
	imgOverlay
		@return: Blended Value
		@desc:	if (dst > ) then: 1 - (1 - 2 * (dst - )) * (1 - src)
			if (dst <= ) then: (2 * dst) * src
]]
function imgOverlay (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( dst.r > 0.5 ) then
		blend.r = 1 - (1 - 2 * (dst.r - 0.5)) * (1 - src.r);
	else
		blend.r = 2 * dst.r * src.r;
	end

	-- Green Blend
	if ( dst.g > 0.5 ) then
		blend.g = 1 - (1 - 2 * (dst.g - 0.5)) * (1 - src.g);
	else
		blend.g = 2 * dst.g * src.g;
	end

	-- Blue Blend
	if ( dst.b > 0.5 ) then
		blend.b = 1 - (1 - 2 * (dst.b - 0.5)) * (1 - src.b);
	else
		blend.b = 2 * dst.b * src.b;
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end

--[[
	imgSoftLight
		@return: Blended Value
		@desc:	if (src > ) then: dst * (1 - (1 - dst) * (1 - 2 * (src - )))
			if (src <= ) then: 1 - (1 - dst) * (1 - (dst * (2 * src))) 
]]
function imgSoftLight (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( src.r > 0.5 ) then
		blend.r = dst.r * (1 - (1 - dst.r) * (1 - 2 * (src.r - 0.5)));
	else
		blend.r = 1 - (1 - dst.r) * (1 - (dst.r * (2 * src.r)));
	end

	-- Green Blend
	if ( src.g > 0.5 ) then
		blend.g = dst.g * (1 - (1 - dst.g) * (1 - 2 * (src.g - 0.5)));
	else
		blend.g = 1 - (1 - dst.g) * (1 - (dst.g * (2 * src.g)));
	end

	-- Blue Blend
	if ( src.b > 0.5 ) then
		blend.b = dst.b * (1 - (1 - dst.b) * (1 - 2 * (src.b - 0.5)));
	else
		blend.b = 1 - (1 - dst.b) * (1 - (dst.b * (2 * src.b)));
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end

--[[
	imgHardLight
		@return: Blended Value
		@desc:	if (src > ) then: 1 - (1 - dst) * (1 - 2 * (src - ))
			if (src <= ) then: dst * (2 * src) 
]]
function imgHardLight (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( src.r > 0.5 ) then
		blend.r = 1 - (1 - dst.r) * (1 - 2 * (src.r - 0.5));
	else
		blend.r = dst.r * (2 * src.r);
	end

	-- Green Blend
	if ( src.g > 0.5 ) then
		blend.g = 1 - (1 - dst.g) * (1 - 2 * (src.g - 0.5));
	else
		blend.g = dst.g * (2 * src.g);
	end

	-- Blue Blend
	if ( src.b > 0.5 ) then
		blend.b = 1 - (1 - dst.b) * (1 - 2 * (src.b - 0.5));
	else
		blend.b = dst.b * (2 * src.b);
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end

--[[
	imgVividLight
		@return: Blended Value
		@desc:	if (src > ) then: 1 - (1 - dst) / (2 * (src - ))
			if (src <= ) then: dst / (1 - 2 * src)
]]
function imgVividLight (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( src.r > 0.5 ) then
		blend.r = 1 - (1 - dst.r) / (2 * (src.r - 0.5));
	else
		blend.r = dst.r / (1 - 2 * src.r);
	end

	-- Green Blend
	if ( src.g > 0.5 ) then
		blend.g = 1 - (1 - dst.g) / (2 * (src.g - 0.5));
	else
		blend.g = dst.g / (1 - 2 * src.g);
	end

	-- Blue Blend
	if ( src.b > 0.5 ) then
		blend.b = 1 - (1 - dst.b) / (2 * (src.b - 0.5));
	else
		blend.b = dst.b / (1 - 2 * src.b);
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end

--[[
	imgLinearLight
		@return: Blended Value
		@desc:	if (src > ) then: = dst + 2 * (src - )
			if (src <= ) then: dst + 2 * src - 1
]]
function imgLinearLight (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( src.r > 0.5 ) then
		blend.r = dst.r + 2 * (src.r - 0.5);
	else
		blend.r = dst.r + 2 * src.r - 1;
	end

	-- Green Blend
	if ( src.g > 0.5 ) then
		blend.g = dst.g + 2 * (src.g - 0.5);
	else
		blend.g = dst.g + 2 * src.g - 1;
	end

	-- Blue Blend
	if ( src.b > 0.5 ) then
		blend.b = dst.b + 2 * (src.b - 0.5);
	else
		blend.b = dst.b + 2 * src.b - 1;
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end

--[[
	imgPinLight
		@return: Blended Value
		@desc:	if (src > ) then: max(dst, 2 * (src - ))
			if (src <= ) then: min(dst, 2 * src))
]]
function imgPinLight (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Red Blend
	if ( src.r > 0.5 ) then
		blend.r = math.max(dst.r, 2 * (src.r - 0.5));
	else
		blend.r = math.min(dst.r, 2 * src.r);
	end

	-- Green Blend
	if ( src.g > 0.5 ) then
		blend.g = math.max(dst.g, 2 * (src.g - 0.5));
	else
		blend.g = math.min(dst.g, 2 * src.g);
	end

	-- Blue Blend
	if ( src.b > 0.5 ) then
		blend.b = math.max(dst.b, 2 * (src.b - 0.5));
	else
		blend.b = math.min(dst.b, 2 * src.b);
	end

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end


--[=[
	Logic Blends
]=]
--[[
	imgDifference
		@return: Blended Value
		@desc: |dst - src|
]]
function imgDifference (dst, src, blend)
	blend.r = math.abs(dst.r - src.r);
	blend.g = math.abs(dst.g - src.g);
	blend.b = math.abs(dst.b - src.b);

	return imgColourCheck(blend);
end

--[[
	imgExclusion
		@return: Blended Value
		@desc:  - 2 * (dst - ) * (src - )
]]
function imgExclusion (dst, src, blend)
	-- Convert colours to floating point
	ColourFloat(dst);
	ColourFloat(src);

	-- Blend
	blend.r = 0.5 - 2 * (dst.r - 0.5) * (src.r - 0.5);
	blend.g = 0.5 - 2 * (dst.g - 0.5) * (src.g - 0.5);
	blend.b = 0.5 - 2 * (dst.b - 0.5) * (src.b - 0.5);

	-- Resample to [0, 255]
	blend.r = blend.r * 255;
	blend.g = blend.g * 255;
	blend.b = blend.b * 255;

	return imgColourCheck(blend);
end
