----------------------------------------------------------
--		Curve Script				--
--							--
--  Author:	Nathaniel Meyer				--
--  Email:	nate@nutty.ca				--
--  Copyright:	Nutty Productions			--
--  Website:	http://www.nutty.ca			--
--							--
--  Desc:	Contains a set of math routines		--
--							--
----------------------------------------------------------


--[[
	Point
		@return: A point structure with (x,y,z) coordinates
]]
function Point (x, y, z)
	return { x = x, y = y, z = z };
end


--[[
	normalize
		@param: point
			Specify a point to be normalized
		@return: Point structure containing the normalized result
]]
function normalize (point)
	local unit = math.sqrt(point.x * point.x + point.y * point.y + point.z * point.z);
	if unit ~= 0 then
		point.x = point.x / unit;
		point.y = point.y / unit;
		point.z = point.z / unit;
	end
end


--[[
	dot
		@param: point1
			Point 1
		@param: point2
			Point 2
		@return: Dot product of point1 and point2
]]
function dot (p1, p2)
	return (p1.x * p2.x + p1.y * p2.y + p1.z * p2.z);
end


--[[
	reflect
		@param: normal
			(x, y, z) normal
		@param: incident
			(x, y, z) incident vector
]]
function reflect (normal, incident)
	-- Calculate the cosine reflectance
	local cosI = 2 * dot(normal, incident);

	-- Adjust the normal
	vec = Point(cosI * normal.x,
		    cosI * normal.y,
		    cosI * normal.z);

	-- Calculate the reflected angle
	return {x = incident.x * vec.x,
		y = incident.y * vec.y,
		z = incident.z * vec.z};
end


--[[
	refract
		@param: normal
			(x, y, z) normal
		@param: incident
			(x, y, z) incident vector
		@param: n1
			Index of the source material (source casting refraction)
		@param: n2
			Index of the destination material
]]
function refract (normal, incident, n1, n2)
	if ( n2 ~= 0 ) then
		-- Get the index ratio
		local n = n1 / n2;

		-- Calculate the cosine reflectance
		local cosI = dot(normal, incident);

		-- Calculate snell's law
		local sinT2 = (n * n) * (1 - (cosI * cosI));

		-- Check for Total Internal Reflectance
		if ( sinT2 <= 1 ) then
			local vec1 = Point(incident.x * n,
					   incident.y * n,
					   incident.z * n);

			local result = n + math.sqrt(1 - sinT2);
			local vec2 = Point(normal.x * result,
					   normal.y * result,
					   normal.z * result);

			return {x = vec1.x - vec2.x,
				y = vec1.y - vec2.y,
				z = vec1.z - vec2.z};
		end
	end

	-- Invalid, return empty
	return {x = 0, y = 0, z = 0};
end
